#!/usr/bin/env python
"""TcEx Framework Package Module."""
# standard library
import fnmatch
import json
import os
import shutil
from pathlib import Path
from typing import List, Optional

# third-party
import colorama as c

# first-party
from tcex.app_config.install_json import InstallJson
from tcex.backports import cached_property
from tcex.bin.bin_abc import BinABC


class Package(BinABC):
    """Package ThreatConnect Job or Playbook App for deployment.

    This method will package the app for deployment to ThreatConnect. Validation of the
    install.json file or files will be automatically run before packaging the app.
    """

    def __init__(self, excludes: Optional[List[str]], ignore_validation: bool, output_dir: Path):
        """Initialize Class properties."""
        super().__init__()
        self._excludes = excludes or []
        self.ignore_validation = ignore_validation
        self.output_dir = output_dir

        # properties
        self.features = ['aotExecutionEnabled', 'secureParams']
        self.package_data = {'errors': [], 'updates': [], 'package': []}
        self.validation_data = {}

    @cached_property
    def _build_excludes_glob(self):  # pylint: disable=no-self-use
        """Return a list of files and folders that should be excluded during the build process."""
        # glob files/directories
        return [
            '__pycache__',
            '.pytest_cache',  # pytest cache directory
            '*.iml',  # PyCharm files
            '*.pyc',  # any pyc file
            '*.zip',  # any zip file
        ]

    @cached_property
    def _build_excludes_base(self):
        """Return a list of files/folders that should be excluded in the App base directory."""
        # base directory files/directories
        excludes = [
            self.output_dir,
            '.cache',  # local cache directory
            '.c9',  # C9 IDE
            '.coverage',  # coverage file
            '.coveragerc',  # coverage configuration file file
            '.cspell',  # cspell configuration file
            '.env',  # local environment file
            '.git',  # git directory
            '.gitignore',  # git ignore file
            '.gitlab-ci.yml',  # gitlab ci file
            '.gitmodules',  # git modules
            '.history',  # vscode history plugin
            '.idea',  # PyCharm
            '.pre-commit-config.yaml',  # pre-commit configuration file
            '.prettierrc.toml',  # prettier configuration file
            '.python-version',  # pyenv
            '.template_manifest.json',  # template manifest file
            '.vscode',  # Visual Studio Code
            'angular.json',  # angular configuration file
            'app.yaml',  # requirements builder configuration file
            'artifacts',  # pytest in CI/CD
            'assets',  # pytest in BB Pipelines
            'cspell.json',  # cspell configuration file
            'local-*',  # log directory
            'log',  # log directory
            'JIRA.html',  # documentation file
            'JIRA.md',  # documentation file
            'karma.conf.js',  # karma configuration file
            'package-lock.json',  # npm package lock file
            'package.json',  # npm package file
            'pyproject.toml',  # project configuration file
            'README.html',  # documentation file
            'target',  # the target directory for builds
            'test-reports',  # pytest in CI/CD
            'tests',  # pytest test directory
        ]
        # excludes.extend(self._build_excludes_base)
        excludes.extend(self._excludes)
        excludes.extend(self.tj.model.package.excludes)
        return excludes

    def exclude_files(self, src: str, names: list):
        """Ignore exclude files in shutil.copytree (callback)."""
        exclude_list = self._build_excludes_glob
        if src == os.getcwd():
            # get excludes that are specific to the Apps base directory
            exclude_list = self._build_excludes_base

        excluded_files = []
        for n in names:
            for e in exclude_list:
                if fnmatch.fnmatch(n, e):
                    excluded_files.append(n)
        return excluded_files

    @cached_property
    def build_fqpn(self) -> 'Path':
        """Return the fully qualified path name of the build directory."""
        build_fqpn = Path(os.path.join(self.app_path, self.output_dir.name, 'build'))
        build_fqpn.mkdir(exist_ok=True, parents=True)
        return build_fqpn

    @cached_property
    def template_fqpn(self) -> 'Path':
        """Return the fully qualified path name of the template directory."""
        template_fqpn = Path(os.path.join(self.build_fqpn, 'template'))
        if os.access(template_fqpn, os.W_OK):
            # cleanup any previous failed builds
            shutil.rmtree(template_fqpn)

        # update package data
        self.package_data['package'].append(
            {'action': 'Template Directory:', 'output': template_fqpn.name}
        )
        return template_fqpn

    def package(self):
        """Build the App package for deployment to ThreatConnect Exchange."""
        # copy project directory to temp location to use as template for multiple builds
        shutil.copytree(self.app_path, self.template_fqpn, False, ignore=self.exclude_files)

        # update package data
        self.package_data['package'].append(
            {'action': 'App Name:', 'output': self.tj.model.package.app_name}
        )

        # use developer defined app version (deprecated) or package_version from InstallJson model
        app_version = self.tj.model.package.app_version or self.ij.model.package_version

        # update package data
        self.package_data['package'].append({'action': 'App Version:', 'output': f'{app_version}'})

        # !!! The name of the folder in the zip is the *key* for an App. This value must
        # !!! remain consistent for the App to upgrade successfully.
        app_name_version = f'{self.tj.model.package.app_name}_{app_version}'

        # build app directory
        app_path_fqpn = os.path.join(self.build_fqpn, app_name_version)
        if os.access(app_path_fqpn, os.W_OK):
            # cleanup any previous failed builds
            shutil.rmtree(app_path_fqpn)
        shutil.copytree(self.template_fqpn, app_path_fqpn)

        # load template install json
        ij_template = InstallJson(path=app_path_fqpn)

        # automatically update install.json in template directory
        ij_template.update.multiple(sequence=False, valid_values=False, playbook_data_types=False)

        # zip file
        self.zip_file(self.app_path, app_name_version, self.build_fqpn)

        # cleanup build directory
        shutil.rmtree(app_path_fqpn)

    def print_json(self):
        """[App Builder] Print JSON output containing results of the package command."""
        print(
            json.dumps({'package_data': self.package_data, 'validation_data': self.validation_data})
        )

    def print_results(self):
        """Print results of the package command."""
        # Updates
        if self.package_data.get('updates'):
            print(f'\n{c.Style.BRIGHT}{c.Fore.BLUE}Updates:')
            for p in self.package_data['updates']:
                print(
                    f"{p.get('action')!s:<20}{c.Style.BRIGHT}{c.Fore.CYAN} {p.get('output')!s:<50}"
                )

        # Packaging
        print(f'\n{c.Style.BRIGHT}{c.Fore.BLUE}Package:')
        for p in self.package_data['package']:
            if isinstance(p.get('output'), list):
                n = 5
                list_data = p.get('output')
                print(
                    f"{p.get('action'):<20}{c.Style.BRIGHT}{c.Fore.CYAN} "
                    f"{', '.join(p.get('output')[:n]):<50}"
                )
                del list_data[:n]
                for data in [
                    list_data[i : i + n] for i in range(0, len(list_data), n)  # noqa: E203
                ]:
                    print(f'''{''!s:<20}{c.Style.BRIGHT}{c.Fore.CYAN} {', '.join(data)!s:<50}''')

            else:
                print(
                    f'''{p.get('action')!s:<20}{c.Style.BRIGHT}'''
                    f'''{c.Fore.CYAN} {p.get('output')!s:<50}'''
                )

        # ignore exit code
        if not self.ignore_validation:
            print('\n')  # separate errors from normal output
            # print all errors
            for error in self.package_data.get('errors'):
                print(f'{c.Fore.RED}{error}')
                self.exit_code = 1

    def zip_file(self, app_path: Path, app_name: Path, tmp_path: Path):
        """Zip the App with tcex extension.

        Args:
            app_path: The path of the current project.
            app_name: The name of the App.
            tmp_path: The temp output path for the zip.
        """
        # zip build directory
        zip_fqpn = Path(os.path.join(app_path, self.output_dir, app_name))

        # create App package
        shutil.make_archive(zip_fqpn, format='zip', root_dir=tmp_path, base_dir=app_name)

        # rename the app swapping .zip for .tcx, some filename have "v1.0" which causes
        # the extra dot to be treated as an extension in pathlib.
        zip_fqfn = os.path.join(app_path, self.output_dir, f'{app_name}.zip')
        tcx_fqfn = os.path.join(app_path, self.output_dir, f'{app_name}.tcx')
        shutil.move(zip_fqfn, tcx_fqfn)

        # update package data
        self.package_data['package'].append({'action': 'App Package:', 'output': tcx_fqfn})



"""Field Types"""

# flake8:noqa
# first-party
from tcex.input.field_types.binary import Binary, binary
from tcex.input.field_types.case_management_entity import (
    ArtifactEntity,
    ArtifactTypeEntity,
    CaseEntity,
    NoteEntity,
    TaskEntity,
    WorkflowEventEntity,
    WorkflowTemplateEntity,
)
from tcex.input.field_types.choice import Choice, choice
from tcex.input.field_types.datetime import DateTime
from tcex.input.field_types.edit_choice import EditChoice, edit_choice
from tcex.input.field_types.group_entity import GroupEntity
from tcex.input.field_types.indicator_entity import (
    AddressEntity,
    EmailAddressEntity,
    FileEntity,
    HostEntity,
    IndicatorEntity,
    UrlEntity,
    indicator_entity,
)
from tcex.input.field_types.integer import Integer, integer
from tcex.input.field_types.ip_address import IpAddress, ip_address
from tcex.input.field_types.key_value import KeyValue
from tcex.input.field_types.sensitive import Sensitive, sensitive
from tcex.input.field_types.string import String, string
from tcex.input.field_types.tc_entity import TCEntity
from tcex.input.field_types.validators import (
    always_array,
    conditional_required,
    entity_input,
    modify_advanced_settings,
)



"""ThreatConnect Requests Session"""
# standard library
import time
from typing import TYPE_CHECKING, Callable, Optional, Union

# first-party
from tcex.sessions.auth.hmac_auth import HmacAuth
from tcex.sessions.auth.token_auth import TokenAuth

if TYPE_CHECKING:  # pragma: no cover
    # first-party
    from tcex.input.field_types.sensitive import Sensitive


class TcAuth(HmacAuth, TokenAuth):
    """ThreatConnect Token Authorization"""

    def __init__(
        self,
        tc_api_access_id: Optional[str] = None,
        tc_api_secret_key: Optional['Sensitive'] = None,
        tc_token: Optional[Union[Callable, str, 'Sensitive']] = None,
    ):
        """Initialize Class Properties."""
        # super(HmacAuth).__init__(tc_api_access_id, tc_api_secret_key)
        # super(TokenAuth, self).__init__(tc_token)
        HmacAuth.__init__(self, tc_api_access_id, tc_api_secret_key)
        TokenAuth.__init__(self, tc_token)

    def __call__(self, r):
        """Add the authorization headers to the request."""
        timestamp = int(time.time())
        if self.tc_api_access_id is not None and self.tc_api_secret_key is not None:
            r.headers['Authorization'] = self._hmac_header(r, timestamp)
        elif self.tc_token is not None:
            r.headers['Authorization'] = self._token_header()
        else:  # pragma: no cover
            raise RuntimeError('No valid ThreatConnect API credentials provided.')

        # Add required headers to auth.
        r.headers['Timestamp'] = timestamp
        return r



"""Proxies"""
# standard library
from typing import TYPE_CHECKING, Optional
from urllib.parse import quote

if TYPE_CHECKING:
    # first-party
    from tcex.input.field_types.sensitive import Sensitive


def proxies(
    proxy_host: Optional[str],
    proxy_port: Optional[int],
    proxy_user: Optional[str],
    proxy_pass: Optional['Sensitive'],
) -> dict:
    """Format the proxy configuration for Python Requests module.

    Generates a dictionary for use with the Python Requests module format
    when proxy is required for remote connections.

    **Example Response**
    ::

        {"http": "http://user:pass@10.10.1.10:3128/"}
    """
    _proxies = {}
    if proxy_host is not None and proxy_port is not None:
        proxy_auth = ''
        if proxy_user is not None and proxy_pass is not None:
            proxy_user = quote(proxy_user, safe='~')
            proxy_pass = quote(proxy_pass.value, safe='~')

            # proxy url with auth
            proxy_auth = f'{proxy_user}:{proxy_pass}@'

        proxy_url = f'{proxy_auth}{proxy_host}:{proxy_port}'
        _proxies = {'http': f'http://{proxy_url}', 'https': f'http://{proxy_url}'}
    return _proxies



"""TcEx Framework Playbook module"""
# standard library
import logging
from typing import Optional, Union

# first-party
from tcex.backports import cached_property
from tcex.key_value_store import KeyValueApi, KeyValueRedis
from tcex.playbook.playbook_create import PlaybookCreate
from tcex.playbook.playbook_delete import PlaybookDelete
from tcex.playbook.playbook_output import PlaybookOutput
from tcex.playbook.playbook_read import PlaybookRead
from tcex.utils.utils import Utils

# get tcex logger
logger = logging.getLogger('tcex')


class Playbook:
    """Playbook methods for accessing key value store.

    Args:
        key_value_store: A KV store instance.
        context: The KV Store context/session_id. For PB Apps the context is provided on
            startup, but for service Apps each request gets a different context.
        output_variables: The requested output variables. For PB Apps outputs are provided on
            startup, but for service Apps each request gets different outputs.
    """

    def __init__(
        self,
        key_value_store: Union[KeyValueApi, KeyValueRedis],
        context: Optional[str] = None,
        output_variables: Optional[list] = None,
    ):
        """Initialize the class properties."""
        self.context = context
        self.key_value_store = key_value_store
        self.output_variables = output_variables or []

        # properties
        self.log = logger
        self.utils = Utils()

    def check_key_requested(self, key: str) -> bool:
        """Return True if output key was requested by downstream app.

        Provide key should be in format "app.output".
        """
        return key in [
            self.utils.get_playbook_variable_model(variable).key
            for variable in self.output_variables
        ]

    def check_variable_requested(self, variable: str) -> bool:
        """Return True if output variable was requested by downstream app.

        Provide variable should be in format of "#App:1234:app.output!String".
        """
        return variable in self.create.output_variables

    def get_variable_type(self, variable: str) -> str:
        """Get the Type from the variable string or default to String type.

        The default type is "String" for those cases when the input variable is
        contains not "DB variable" and is just a String.

        Example Variable:

        #App:1234:output!StringArray returns **StringArray**

        Example String:

        "My Data" returns **String**
        """
        return self.utils.get_playbook_variable_type(variable)

    @cached_property
    def create(self) -> 'PlaybookCreate':
        """Return instance of PlaybookCreate"""
        return PlaybookCreate(self.context, self.key_value_store, self.output_variables)

    @cached_property
    def delete(self) -> 'PlaybookDelete':
        """Return instance of PlaybookDelete"""
        return PlaybookDelete(self.context, self.key_value_store)

    def is_variable(self, key: str) -> bool:
        """Return True if provided key is a properly formatted playbook variable."""
        return self.utils.is_playbook_variable(key)

    @cached_property
    def output(self) -> 'PlaybookOutput':
        """Return instance of PlaybookOutput"""
        return PlaybookOutput(self)

    @cached_property
    def read(self) -> 'PlaybookRead':
        """Return instance of PlaybookRead"""
        return PlaybookRead(self.context, self.key_value_store)



#!/usr/bin/env python
"""TcEx Dependencies Command"""
# standard library
import os
import sys
from pathlib import Path

# third-party
import typer

# first-party
from tcex.bin.bin_abc import BinABC
from tcex.input.field_types.sensitive import Sensitive
from tcex.pleb.proxies import proxies
from tcex.sessions.auth.hmac_auth import HmacAuth
from tcex.sessions.tc_session import TcSession


class Deploy(BinABC):
    """Install dependencies for App."""

    def __init__(
        self,
        server: str,
        allow_all_orgs: bool,
        allow_distribution: bool,
        app_file: str,
        proxy_host: str,
        proxy_port: int,
        proxy_user: str,
        proxy_pass: str,
    ):
        """Initialize Class properties."""
        super().__init__()
        self._app_file = app_file
        self.allow_all_orgs = allow_all_orgs
        self.allow_distribution = allow_distribution
        self.proxy_host = proxy_host
        self.proxy_port = proxy_port
        self.proxy_user = proxy_user
        self.proxy_pass = proxy_pass
        self.server = server

    @staticmethod
    def _handle_missing_environment(variable: str):
        """Print error on missing environment variable."""
        typer.secho(
            f'Could not find environment variable: {variable}.',
            fg=typer.colors.RED,
            err=True,
        )
        sys.exit(1)

    def _check_file(self):
        """Return True if file exists."""
        if not os.path.isfile(self._app_file):
            self.print_failure(f'Could not find file: {self._app_file}.')
            sys.exit(1)

    @property
    def app_file(self):
        """Return app file."""
        if self._app_file is None:
            target_fqpn = Path('target')
            app_files = list(target_fqpn.glob('*.tcx'))
            if len(app_files) > 1:
                self.print_failure(
                    '''More than one App file found, can't autodetect the correct file.''',
                )
                sys.exit(1)
            elif not app_files:
                self.print_failure('No App file found.')

            # set app_file to the only file found
            self._app_file = app_files[0]

        # validate the file exists
        self._check_file()

        return Path(self._app_file)

    @property
    def auth(self):
        """Authenticate with TcEx."""
        tc_api_access_id = os.getenv('TC_API_ACCESS_ID')
        tc_api_secret_key = os.getenv('TC_API_SECRET_KEY')
        if tc_api_access_id is None:
            self._handle_missing_environment('TC_API_ACCESS_ID')
        if tc_api_secret_key is None:
            self._handle_missing_environment('TC_API_SECRET_KEY')

        return HmacAuth(tc_api_access_id, Sensitive(tc_api_secret_key))

    @property
    def base_url(self):
        """Authenticate with TcEx."""
        return f'https://{self.server}/api'

    # pylint: disable=consider-using-with
    def deploy_app(self):
        """Deploy the App to ThreatConnect Exchange."""
        self.print_block('\nUploading App ...\n', fg_color='green')
        files = {
            'allowAllOrgs': self.allow_all_orgs,
            'allowAppDistribution': self.allow_distribution,
            'fileData': ('filename', open(self.app_file, 'rb'), 'application/octet-stream'),
        }
        try:
            response = self.session.post('/internal/apps/exchange/install', files=files, timeout=60)
        except Exception as err:
            self.print_title('Failed Deploying App', fg_color='red')
            self.print_block(str(err), max_length=120)
            sys.exit(0)

        if not response.ok:
            reason = response.text or response.reason
            self.print_title('Failed Deploying App', fg_color='red')
            self.print_setting('File Name', os.path.basename(self.app_file))
            self.print_setting('Reason', reason)
            self.print_setting('Status Code', response.status_code)
            self.print_setting('URL', response.request.url)
        else:
            response_data = response.json()[0]
            self.print_title('Successfully Deployed App')
            self.print_setting('File Name', os.path.basename(self.app_file))
            self.print_setting('Display Name', response_data.get('displayName'))
            self.print_setting('Program Name', response_data.get('programName'))
            self.print_setting('Program Version', response_data.get('programVersion'))
            self.print_setting('Allow All Orgs', self.allow_all_orgs)
            self.print_setting('Allow Distribution', self.allow_distribution)
            self.print_setting('Status Code', response.status_code)
            self.print_setting('URL', response.request.url)

    @property
    def session(self):
        """Create a TcEx Session."""
        _proxies = proxies(
            proxy_host=self.proxy_host,
            proxy_port=self.proxy_port,
            proxy_user=self.proxy_user,
            proxy_pass=self.proxy_pass,
        )
        return TcSession(auth=self.auth, base_url=self.base_url, proxies=_proxies)



"""Input for Apps"""
# standard library
import json
import logging
import os
import re
from base64 import b64decode
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Optional, Union

# third-party
from pydantic import BaseModel, Extra

# first-party
from tcex.app_config.install_json import InstallJson
from tcex.backports import cached_property
from tcex.input.field_types import Sensitive
from tcex.input.models import feature_map, runtime_level_map, tc_action_map
from tcex.key_value_store import RedisClient
from tcex.pleb.none_model import NoneModel
from tcex.pleb.registry import registry
from tcex.utils import Utils

if TYPE_CHECKING:
    # third-party
    from pydantic import ValidationError

# get tcex logger
logger = logging.getLogger('tcex')

# define JSON encoders
json_encoders = {Sensitive: lambda v: str(v)}  # pylint: disable=W0108


def input_model(models: list) -> 'BaseModel':
    """Return Input Model."""

    class InputModel(*models):
        """Input Model"""

        # [legacy] if True, the App should get it's inputs from secure params (redis)
        # supported runtimeLevel: [Organization, Playbook]
        tc_secure_params: bool = False

        # the user id of the one executing the App
        # supported runtimeLevel: [Organization, Playbook]
        tc_user_id: Optional[int]

        class Config:
            """DataModel Config"""

            extra = Extra.allow
            validate_assignment = True
            json_encoders = json_encoders

    return InputModel


class Input:
    """Module to handle inputs for all App types."""

    def __init__(self, config: Optional[dict] = None, config_file: Optional[str] = None, **kwargs):
        """Initialize class properties.

        Keyword Args:
            tc_session: pass a tc_session object to use, else will use the one from registry.
        """

        # TODO [HIGH] kwarg - don't add built-in models, supply custom TcSession object
        self.config = config
        self.config_file = config_file

        # properties
        self._models = []
        self.ij = InstallJson()
        self.log = logger
        self.utils = Utils()
        self.tc_session = kwargs.get('tc_session')

    @staticmethod
    def _get_redis_client(host, port, db):
        """Return RedisClient client"""
        return RedisClient(host=host, port=port, db=db).client

    def _load_aot_params(
        self,
        tc_aot_enabled: bool,
        tc_kvstore_type: str,
        tc_kvstore_host: str,
        tc_kvstore_port: int,
        tc_action_channel: str,
        tc_terminate_seconds: int,
    ) -> Dict[str, any]:
        """Subscribe to AOT action channel."""
        params = {}
        if tc_aot_enabled is not True:
            return params

        if tc_kvstore_type == 'Redis':

            # get an instance of redis client
            redis_client = self._get_redis_client(
                host=tc_kvstore_host,
                port=tc_kvstore_port,
                db=0,
            )

            try:
                self.log.info('feature=inputs, event=blocking-for-aot')
                msg_data = redis_client.blpop(
                    keys=tc_action_channel,
                    timeout=int(tc_terminate_seconds),
                )

                if msg_data is None:  # pragma: no cover
                    # send exit to tcex.exit method
                    registry.ExitService.exit_aot_terminate(
                        code=1, msg='AOT subscription timeout reached.'
                    )

                msg_data = json.loads(msg_data[1])
                msg_type = msg_data.get('type', 'terminate')
                if msg_type == 'execute':
                    params = msg_data.get('params', {})
                elif msg_type == 'terminate':
                    # send exit to tcex.exit method
                    registry.ExitService.exit_aot_terminate(
                        code=0, msg='Received AOT terminate message.'
                    )
            except Exception as e:  # pragma: no cover
                # send exit to tcex.exit method
                registry.ExitService.exit_aot_terminate(
                    code=1, msg=f'Exception during AOT subscription ({e}).'
                )

        return params

    def _load_config_file(self):
        """Load config file params provided passed to inputs."""
        # default file contents
        file_content = {}
        if self.config_file is None:
            return file_content

        # self.config_file should be a fully qualified file name
        fqfn = Path(self.config_file)
        if not fqfn.is_file():
            self.log.error(
                'feature=inputs, event=load-config-file, '
                f'exception=file-not-found, filename={fqfn.name}'
            )
            return file_content

        # read file contents
        try:
            # read encrypted file from "in" directory
            with fqfn.open(mode='rb') as fh:
                return json.load(fh)
        except Exception:  # pragma: no cover
            self.log.error(f'feature=inputs, event=config-parse-failure, filename={fqfn.name}')
            return file_content

    def _load_file_params(self):
        """Load file params provided by the core platform."""
        # default file contents
        file_content = {}

        tc_app_param_file = os.getenv('TC_APP_PARAM_FILE')
        tc_app_param_key = os.getenv('TC_APP_PARAM_KEY')
        if not all([tc_app_param_file, tc_app_param_key]):
            return file_content

        # tc_app_param_file is a fully qualified file name
        fqfn = Path(tc_app_param_file)
        if not fqfn.is_file():  # pragma: no cover
            self.log.error(
                'feature=inputs, event=load-file-params, '
                f'exception=file-not-found, filename={fqfn.name}'
            )
            return file_content

        # read file contents
        try:
            # read encrypted file from "in" directory
            with fqfn.open(mode='rb') as fh:
                encrypted_contents = fh.read()
        except Exception:  # pragma: no cover
            self.log.error(f'feature=inputs, event=config-parse-failure, filename={fqfn.name}')
            return file_content

        # decrypt file contents
        try:
            file_content = json.loads(
                self.utils.decrypt_aes_cbc(tc_app_param_key, encrypted_contents).decode()
            )

            # delete file
            fqfn.unlink()
        except Exception:  # pragma: no cover
            self.log.error(f'feature=inputs, event=config-decryption-failure, filename={fqfn.name}')

        return file_content

    def add_model(self, model: BaseModel):
        """Add additional input models."""
        if model:
            self._models.insert(0, model)

        # clear cache for data property
        if 'model' in self.__dict__:
            del self.__dict__['model']

        # add App level models based on special "tc_action" input
        if hasattr(self.model_unresolved, 'tc_action'):
            self._models.extend(tc_action_map.get(self.model_unresolved.tc_action, []))

        # force data model to load so that validation is done at this EXACT point
        _ = self.model

    @cached_property
    def contents(self) -> dict:
        """Return contents of inputs from all locations."""
        _contents = {}

        # config
        if isinstance(self.config, dict):
            _contents.update(self.config)

        # config file
        _contents.update(self._load_config_file())

        # file params
        _contents.update(self._load_file_params())

        # aot params - must be loaded last so that it has the kv store channels
        _contents.update(
            self._load_aot_params(
                tc_aot_enabled=self.utils.to_bool(_contents.get('tc_aot_enabled', False)),
                tc_kvstore_type=_contents.get('tc_kvstore_type'),
                tc_kvstore_host=_contents.get('tc_kvstore_host'),
                tc_kvstore_port=_contents.get('tc_kvstore_port'),
                tc_action_channel=_contents.get('tc_action_channel'),
                tc_terminate_seconds=_contents.get('tc_terminate_seconds'),
            )
        )
        return _contents

    @cached_property
    def contents_resolved(self) -> dict:
        """Resolve all file, keychain, playbook, and text variables.

        Job, Playbook, and Service Apps call can have a tc-variable, but only
        Playbook Apps will have a playbook variable.
        """
        _inputs = self.contents.copy()

        # support external Apps that don't have an install.json
        if not self.ij.fqfn.is_file():  # pragma: no cover
            return _inputs

        for name, value in _inputs.items():
            if name == 'tc_playbook_out_variables':
                # for services, this input contains the name of the expected outputs.  If we don't
                # skip this, we'll try to resolve the value (e.g.
                # #Trigger:334:example.service_input!String), but that 1) won't work for services
                # and 2) doesn't make sense.  Service configs will never have playbook variables.
                continue

            if self.utils.is_tc_variable(value):  # only matches playbook variables
                value = self.resolve_variable(variable=value)
            elif self.ij.model.is_playbook_app:
                if isinstance(value, list):
                    # list could contain playbook variables, try to resolve the value
                    updated_value_array = []
                    for v in value:
                        if isinstance(v, str):
                            v = registry.playbook.read.variable(v)
                        # TODO: [high] does resolve variable need to be added here
                        updated_value_array.append(v)
                    value = updated_value_array
                elif self.utils.is_playbook_variable(value):  # only matches playbook variables
                    # when using Union[Bytes, String] in App input model the value
                    # can be coerced to the wrong type. the BinaryVariable and
                    # StringVariable custom types allows for the validator in Binary
                    # and String types to raise a value error.
                    value = registry.playbook.read.variable(value)
                elif isinstance(value, str):
                    value = registry.playbook.read._read_embedded(value)
            else:
                for match in re.finditer(self.utils.variable_tc_pattern, str(value)):
                    variable = match.group(0)  # the full variable pattern
                    if match.group('type').lower() == 'file':
                        v = '<file>'
                    else:
                        v = self.resolve_variable(variable=variable)
                    value = value.replace(variable, v)

            _inputs[name] = value

        # update contents
        self.contents_update(_inputs)
        return dict(sorted(_inputs.items()))

    # TODO: [high] - can this be replaced with a pydantic root validator?
    def contents_update(self, inputs: dict):
        """Update inputs provided by AOT to be of the proper value and type."""
        for name, value in inputs.items():
            # ThreatConnect AOT params could be updated in the future to proper JSON format.
            # MultiChoice data should be represented as JSON array and Boolean values should be a
            # JSON boolean and not a string.
            param = self.ij.model.get_param(name)
            if isinstance(param, NoneModel):
                # skip over "default" inputs not defined in the install.json file
                continue

            if param.type.lower() == 'multichoice' or param.allow_multiple:
                # update delimited value to an array for inputs that have type of MultiChoice
                if value is not None and not isinstance(value, list):
                    inputs[name] = value.split(self.ij.model.list_delimiter or '|')
            elif param.type == 'boolean' and isinstance(value, str):
                # convert boolean input that are passed in as a string ("true" -> True)
                inputs[name] = value.lower() == 'true'

    @cached_property
    def model(self) -> 'BaseModel':
        """Return the Input Model."""
        return input_model(self.models)(**self.contents_resolved)

    @cached_property
    def model_unresolved(self) -> 'BaseModel':
        """Return the Input Model using contents (no resolved values)."""
        return input_model(self.models)(**self.contents)

    @cached_property
    def models(self) -> list:
        """Return all models for inputs."""
        # support external Apps that don't have an install.json
        if not self.ij.fqfn.is_file():
            return runtime_level_map.get('external')

        # add all models for any supported features of the App
        for feature in self.ij.model.features:
            self._models.extend(feature_map.get(feature, []))

        # add all models based on the runtime level of the App
        self._models.extend(runtime_level_map.get(self.ij.model.runtime_level.lower()))

        return self._models

    # TODO: [low] is this needed or can Field value be set to tc_property=True?
    @cached_property
    def model_properties(self) -> set:
        """Return only defined properties from model (exclude additional)."""
        properties = set()
        for model in self.models:
            properties.update(model.schema().get('properties').keys())

        return properties

    def resolve_variable(self, variable: str) -> Union[bytes, str]:
        """Resolve FILE/KEYCHAIN/TEXT variables.

        Feature: PLAT-2688

        Data Format:
        {
            "data": "value"
        }
        """
        match = re.match(Utils().variable_tc_match, variable)
        key = match.group('key')
        provider = match.group('provider')
        type_ = match.group('type')

        # retrieve value from API
        data = None
        session = self.tc_session if self.tc_session else registry.session_tc
        r = session.get(f'/internal/variable/runtime/{provider}/{key}')
        if r.ok:
            try:
                data = r.json().get('data')

                if type_.lower() == 'file':
                    data = b64decode(data)  # returns bytes
                elif type_.lower() == 'keychain':
                    data = Sensitive(data)
            except Exception as ex:
                raise RuntimeError(
                    f'Could not retrieve variable: provider={provider}, key={key}, type={type_}.'
                ) from ex
        else:
            raise RuntimeError(
                f'Could not retrieve variable: provider={provider}, key={key}, type={type_}.'
            )

        return data

    @staticmethod
    def validation_exit_message(ex: 'ValidationError'):
        """Format and return validation error message."""
        _exit_message = {}
        for err in ex.errors():
            # deduplicate error messagese
            err_loc = ','.join([str(e) for e in err.get('loc')])
            err_msg = err.get('msg')
            _exit_message.setdefault(err_loc, [])
            if err_msg not in _exit_message[err_loc]:
                _exit_message[err_loc].append(err_msg)

        # format error messages
        _exit_message_list = ['Input validation errors']
        for loc, err_list in _exit_message.items():
            _exit_message_list.append(f'''{loc}: {', '.join(err_list)}''')

        # exit with error message
        return '\n'.join(_exit_message_list)



#!/usr/bin/env python
"""TcEx Framework Bin Command Base Module."""
# standard library
import logging
import os
import sys
from abc import ABC
from pathlib import Path
from typing import Optional

# third-party
import typer
from click import Choice

# first-party
from tcex.app_config.install_json import InstallJson
from tcex.app_config.layout_json import LayoutJson
from tcex.app_config.tcex_json import TcexJson
from tcex.backports import cached_property
from tcex.logger.rotating_file_handler_custom import (  # pylint: disable=no-name-in-module
    RotatingFileHandlerCustom,
)
from tcex.logger.trace_logger import TraceLogger  # pylint: disable=no-name-in-module
from tcex.utils.utils import Utils


class BinABC(ABC):
    """Base Class for ThreatConnect command line tools."""

    def __init__(self):
        """Initialize Class properties."""
        # properties
        self.app_path = os.getcwd()
        self.exit_code = 0
        self.i1 = ' ' * 4  # indent level 1
        self.i2 = ' ' * 8  # indent level 2
        self.i3 = ' ' * 12  # indent level 3
        self.ij = InstallJson()
        self.lj = LayoutJson()
        self.tj = TcexJson()
        self.utils = Utils()

    @cached_property
    def cli_out_path(self) -> 'Path':  # pylint: disable=no-self-use
        """Return the path to the tcex cli command out directory."""
        _out_path = Path(os.path.expanduser('~/.tcex'))
        _out_path.mkdir(exist_ok=True, parents=True)
        return _out_path

    @staticmethod
    def handle_error(err, halt: Optional[bool] = True):
        """Print errors message and optionally exit.

        Args:
            err (str): The error message to print.
            halt (bool, optional): Defaults to True. If True the script will exit.
        """
        typer.secho(err, fg=typer.colors.RED, err=True)
        if halt:
            sys.exit(1)

    @cached_property
    def log(self) -> 'logging.Logger':
        """Return the configured logger."""
        # create logger based on custom TestLogger
        logging.setLoggerClass(TraceLogger)

        # init logger
        logger = logging.getLogger('tcex-cli')

        # set logger level
        logger.setLevel(logging.TRACE)  # pylint: disable=no-member

        # create rotation filehandler
        lfh = RotatingFileHandlerCustom(
            backupCount=3,
            filename=f'{self.cli_out_path}/tcex.log',
            maxBytes=100_000,
        )

        # get logging level from OS env or default to debug
        logging_level = logging.getLevelName('DEBUG')

        # set handler logging level
        lfh.setLevel(logging_level)

        # create formatter
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s ' '(%(filename)s:%(lineno)d)'
        )
        if logging_level < 10:
            formatter = logging.Formatter(
                '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
                '(%(filename)s:%(funcName)s:%(lineno)d:%(threadName)s)'
            )

        # set formatter
        lfh.setFormatter(formatter)

        # add handler
        logger.addHandler(lfh)

        return logger

    @staticmethod
    def print_block(text: str, max_length: Optional[int] = 80, **kwargs):
        """Print Divider."""
        bold = kwargs.get('bold', False)
        fg_color = getattr(typer.colors, kwargs.get('fg_color', 'white').upper())

        # split text
        text_wrapped = ''
        for word in text.split(' '):
            if len(text_wrapped) + len(word) < max_length:
                text_wrapped += f'{word} '
            else:
                typer.secho(text_wrapped, fg=fg_color, bold=bold)
                text_wrapped = f'{word} '
        typer.secho(text_wrapped, fg=fg_color, bold=bold)

    @staticmethod
    def print_divider(char: Optional[str] = '-', count: Optional[int] = 100, **kwargs):
        """Print Divider."""
        bold = kwargs.get('bold', False)
        fg_color = getattr(typer.colors, kwargs.get('fg_color', 'bright_white').upper())

        # print divider
        typer.secho(char * count, fg=fg_color, bold=bold)

    @staticmethod
    def print_failure(message: str, exit_: Optional[bool] = True):
        """Print Failure."""
        typer.secho(message, fg=typer.colors.RED, bold=True)
        if exit_ is True:
            sys.exit(1)

    @staticmethod
    def print_setting(label: str, value: str, **kwargs):
        """Print Setting."""
        bold = kwargs.get('bold', True)
        fg_color = getattr(typer.colors, kwargs.get('fg_color', 'magenta').upper())
        indent = ' ' * kwargs.get('indent', 0)

        # print setting
        value_display = typer.style(f'{value}', fg=fg_color, bold=bold)
        typer.echo(f'{indent}{label:<20}: {value_display}')

    @staticmethod
    def print_title(title: str, divider: Optional[bool] = True, **kwargs):
        """Print Title."""
        bold = kwargs.get('bold', True)
        fg_color = getattr(typer.colors, kwargs.get('fg_color', 'cyan').upper())

        # print title
        typer.secho(title, fg=fg_color, bold=bold)
        if divider is True:
            typer.secho('=' * len(title), fg=fg_color, bold=bold)

    @staticmethod
    def prompt_choice(text: str, choices: list, default: str, **kwargs) -> bool:
        """Present a prompt with a bool response."""
        bold = kwargs.get('bold', True)
        fg_color = getattr(typer.colors, kwargs.get('fg_color', 'cyan').upper())

        text = typer.style(f'{text}', fg=fg_color, bold=bold)
        choice = Choice(choices)

        return typer.prompt(
            text=text,
            default=default,
            type=choice,
        )

    @staticmethod
    def update_system_path():
        """Update the system path to ensure project modules and dependencies can be found."""
        cwd = os.getcwd()
        lib_dir = os.path.join(os.getcwd(), 'lib_')
        lib_latest = os.path.join(os.getcwd(), 'lib_latest')

        # insert the lib_latest directory into the system Path if no other lib directory found. This
        # entry will be bumped to index 1 after adding the current working directory.
        if not [p for p in sys.path if lib_dir in p]:
            sys.path.insert(0, lib_latest)

        # insert the current working directory into the system Path for the App, ensuring that it is
        # always the first entry in the list.
        try:
            sys.path.remove(cwd)
        except ValueError:
            pass
        sys.path.insert(0, cwd)



"""Datetime (Arrow) Field Type"""
# standard library
from typing import TYPE_CHECKING, Any, Callable

# third-party
import arrow

# first-party
from tcex.input.field_types.exception import InvalidInput
from tcex.utils import Utils

if TYPE_CHECKING:  # pragma: no cover
    # third-party
    from pydantic.fields import ModelField


class DateTime(arrow.Arrow):
    """Datetime (Arrow) Field Type"""

    @classmethod
    def __get_validators__(cls) -> Callable:
        """Define one or more validators for Pydantic custom type."""
        yield cls._validate

    @classmethod
    def _validate(cls, value: Any, field: 'ModelField') -> 'arrow.Arrow':
        """Pydantic validate method."""
        try:
            return Utils.any_to_datetime(value)
        except RuntimeError as ex:
            raise InvalidInput(field_name=field.name, error=str(ex)) from ex



"""Setup for TcEx Module."""
# standard library
import os
import sys

# third-party
from setuptools import find_packages, setup

metadata = {}
metadata_file = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'tcex', '__metadata__.py')
with open(
    metadata_file,
    encoding='utf-8',
) as f:
    exec(f.read(), metadata)  # nosec; pylint: disable=exec-used

if not metadata:
    raise RuntimeError(f'Could not load metadata file ({metadata_file}).')

with open('README.md') as f:
    readme = f.read()

dev_packages = [
    'black',
    'codespell',
    'fakeredis==1.7.0',
    'flake8',
    'isort>=5.0.0',
    'mako',
    'pre-commit',
    'pydocstyle',
    'pylint>=2.5.0,<2.14.0',
    'pytest',
    'pytest-cov',
    'pytest-html',
    'pytest-ordering',
    'pytest-xdist>=2.5.0',
    'pyupgrade',
]
if sys.version_info <= (3, 7):
    # these packages dropped support for 3.6
    dev_packages.extend(['bandit==1.7.1', 'deepdiff==5.7.0'])
else:
    dev_packages.extend(['bandit', 'deepdiff'])


setup(
    author=metadata['__author__'],
    author_email=metadata['__author_email__'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Security',
    ],
    description=metadata['__description__'],
    download_url=metadata['__download_url__'],
    extras_require={'dev': dev_packages, 'develop': dev_packages, 'development': dev_packages},
    include_package_data=True,
    install_requires=[
        'arrow',
        'astunparse',
        'backports.cached-property; python_version < "3.8.0"',
        'colorama>=0.3.9',
        'inflect>=0.2.5',
        'jmespath',
        'paho-mqtt',
        'pyaes',
        'pydantic',
        'python-dateutil>=2.6.1',
        'pyyaml',
        'redis>=2.10.6',
        'requests',
        'semantic_version',
        'stdlib-list>=0.6.0',
        'tinydb',
        'typer',
        'wrapt',
    ],
    license=metadata['__license__'],
    long_description=readme,
    long_description_content_type='text/markdown',
    name=metadata['__package_name__'],
    packages=find_packages(exclude=['tests', 'tests.*']),
    package_dir={'tcex': 'tcex'},
    project_urls={
        'Documentation': 'https://github.com/ThreatConnect-Inc/tcex',
        'Source': 'https://github.com/ThreatConnect-Inc/tcex',
    },
    python_requires='>=3.6',
    scripts=[
        'bin/tcex',
    ],
    url=metadata['__url__'],
    version=metadata['__version__'],
    zip_safe=True,
)



"""TCEntity Playbook Type"""
# standard library
from typing import TYPE_CHECKING, Dict, Optional

# third-party
from pydantic import BaseModel, Extra, validator

# first-party
from tcex.input.field_types.exception import InvalidEmptyValue

if TYPE_CHECKING:  # pragma: no cover
    # third-party
    from pydantic.fields import ModelField


# pylint: disable=no-self-argument, no-self-use
class TCEntity(BaseModel):
    """Model for TCEntity Input."""

    id: int
    type: str
    value: str

    # IMPORTANT: confidence and rating values are included only so that, when defined,
    #    they come back as the correct types. They are only intended for Indicator
    #    types and are not guaranteed to be populated.
    confidence: Optional[int]
    rating: Optional[int]

    @validator('id', 'type', 'value')
    def non_empty_string(cls, v: Dict[str, str], field: 'ModelField') -> Dict[str, str]:
        """Validate that the value is a non-empty string."""
        if isinstance(v, str) and v.replace(' ', '') == '':  # None value are automatically covered
            raise InvalidEmptyValue(field_name=field.name)
        return v

    class Config:
        """Model Config"""

        extra = Extra.allow
        validate_assignment = True



"""App Spec Yml App Config"""
# standard library
import json
import logging
import os
from pathlib import Path
from typing import Optional

# third-party
import yaml

try:
    # third-party
    from yaml import CDumper as Dumper
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader, Dumper

# first-party
from tcex.app_config.install_json import InstallJson
from tcex.app_config.models import AppSpecYmlModel
from tcex.app_config.tcex_json import TcexJson
from tcex.backports import cached_property

# get tcex logger
tcex_logger = logging.getLogger('tcex')


class AppSpecYml:
    """Provide a model for the app_spec.yml config file."""

    def __init__(
        self,
        filename: Optional[str] = None,
        path: Optional[str] = None,
        logger: Optional[logging.Logger] = None,
    ):
        """Initialize class properties."""
        filename = filename or 'app_spec.yml'
        path = path or os.getcwd()
        self.log = logger or tcex_logger

        # properties
        self.fqfn = Path(os.path.join(path, filename))
        self.ij = InstallJson(logger=self.log)
        self.tj = TcexJson(logger=self.log)

    @property
    def _feature_data_advanced_request_inputs(self):
        """Return all inputs for advanced request."""
        return [
            {
                'display': '''tc_action in ('Advanced Request')''',
                'label': 'API Endpoint/Path',
                'name': 'tc_adv_req_path',
                'note': 'The API Path request.',
                'playbookDataType': ['String'],
                'required': True,
                'type': 'String',
                'validValues': ['${TEXT}'],
            },
            {
                'display': '''tc_action in ('Advanced Request')''',
                'default': 'GET',
                'label': 'HTTP Method',
                'name': 'tc_adv_req_http_method',
                'note': 'HTTP method to use.',
                'required': True,
                'type': 'Choice',
                'validValues': ['GET', 'POST', 'DELETE', 'PUT', 'HEAD', 'PATCH', 'OPTIONS'],
            },
            {
                'display': '''tc_action in ('Advanced Request')''',
                'label': 'Query Parameters',
                'name': 'tc_adv_req_params',
                'note': (
                    'Query parameters to append to the URL. For sensitive information like API '
                    'keys, using variables is recommended to ensure that the Playbook will not '
                    'export sensitive data.'
                ),
                'playbookDataType': ['String', 'StringArray'],
                'required': False,
                'type': 'KeyValueList',
                'validValues': ['${KEYCHAIN}', '${TEXT}'],
            },
            {
                'display': '''tc_action in ('Advanced Request')''',
                'label': 'Exclude Empty/Null Parameters',
                'name': 'tc_adv_req_exclude_null_params',
                'note': (
                    '''Some API endpoint don't handle null/empty query parameters properly '''
                    '''(e.g., ?name=&type=String). If selected this options will exclude any '''
                    '''query parameters that has a null/empty value.'''
                ),
                'type': 'Boolean',
            },
            {
                'display': '''tc_action in ('Advanced Request')''',
                'label': 'Headers',
                'name': 'tc_adv_req_headers',
                'note': (
                    'Headers to include in the request. When using Multi-part Form/File data, '
                    'do **not** add a **Content-Type** header. For sensitive information like '
                    'API keys, using variables is recommended to ensure that the Playbook will '
                    'not export sensitive data.'
                ),
                'playbookDataType': ['String'],
                'required': False,
                'type': 'KeyValueList',
                'validValues': ['${KEYCHAIN}', '${TEXT}'],
            },
            {
                'display': (
                    '''tc_action in ('Advanced Request') AND tc_adv_req_http_method '''
                    '''in ('POST', 'PUT', 'DELETE', 'PATCH')'''
                ),
                'label': 'Body',
                'name': 'tc_adv_req_body',
                'note': 'Content of the HTTP request.',
                'playbookDataType': ['String', 'Binary'],
                'required': False,
                'type': 'String',
                'validValues': ['${KEYCHAIN}', '${TEXT}'],
                'viewRows': 4,
            },
            {
                'display': (
                    '''tc_action in ('Advanced Request') AND tc_adv_req_http_method '''
                    '''in ('POST', 'PUT', 'DELETE', 'PATCH')'''
                ),
                'label': 'URL Encode JSON Body',
                'name': 'tc_adv_req_urlencode_body',
                'note': (
                    '''URL encode a JSON-formatted body. Typically used for'''
                    ''' 'x-www-form-urlencoded' data, where the data can be configured in the '''
                    '''body as a JSON string.'''
                ),
                'type': 'Boolean',
            },
            {
                'display': '''tc_action in ('Advanced Request')''',
                'default': True,
                'label': 'Fail for Status',
                'name': 'tc_adv_req_fail_on_error',
                'note': 'Fail if the response status code is 4XX - 5XX.',
                'type': 'Boolean',
            },
        ]

    @staticmethod
    def _feature_data_advanced_request_outputs(prefix: str) -> dict:
        """Return all outputs for advanced request."""
        return {
            'display': 'tc_action in (\'Advanced Request\')',
            'outputVariables': [
                {
                    'name': f'{prefix}.request.content',
                },
                {
                    'name': f'{prefix}.request.content.binary',
                    'type': 'Binary',
                },
                {
                    'name': f'{prefix}.request.headers',
                },
                {
                    'name': f'{prefix}.request.ok',
                },
                {
                    'name': f'{prefix}.request.reason',
                },
                {
                    'name': f'{prefix}.request.status_code',
                },
                {
                    'name': f'{prefix}.request.url',
                },
            ],
        }

    def _migrate_schema_100_to_110(self, contents: dict):
        """Migrate 1.0.0 schema to 1.1.0 schema."""
        # moved app.* to top level
        self._migrate_schema_100_to_110_app(contents)

        # migrate app.feeds to app.organization.feeds
        self._migrate_schema_100_to_110_organization_feeds(contents)

        # migrate app.feeds to app.organization.repeating_minutes
        self._migrate_schema_100_to_110_organization_repeating_minutes(contents)

        # migrate app.feeds to app.organization.publish_out_files
        self._migrate_schema_100_to_110_organization_publish_out_files(contents)

        # migrate app.inputGroup to new schema
        self._migrate_schema_100_to_110_input_groups(contents)

        # migrate app.note to app.notePerAction with new schema
        self._migrate_schema_100_to_110_notes_per_action(contents)

        # migrate app.outputGroups to outputData with new schema
        self._migrate_schema_100_to_110_output_groups(contents)

        # migrate app.playbookType to category
        contents['category'] = contents.pop('playbookType', '')

        # migrate app.jira to internalNotes schema
        self._migrate_schema_100_to_110_jira_notes(contents)

        # migrate app.releaseNotes to new schema
        self._migrate_schema_100_to_110_release_notes(contents)

        # migrate app.retry to playbook.retry
        self._migrate_schema_100_to_110_retry(contents)

        # update the schema version
        contents['schemaVersion'] = contents.get('schemaVersion') or '1.1.0'

        # dict -> mode -> dict (filtered)
        self.rewrite_contents(contents)

    @staticmethod
    def _migrate_schema_100_to_110_app(contents: dict):
        """Migrate 1.0.0 schema to 1.1.0 schema."""
        # remove "app" top level
        for k, v in dict(contents).get('app').items():
            contents[k] = v

        # assure minServerVersion exists
        if contents.get('minServerVersion') is None:
            contents['minServerVersion'] = '6.0.0'

        # remove "app" from "app_spec"
        del contents['app']

    @staticmethod
    def _migrate_schema_100_to_110_organization_feeds(contents: dict):
        """Migrate 1.0.0 schema to 1.1.0 schema."""
        if contents.get('feeds') is not None and contents['runtimeLevel'].lower() == 'organization':
            contents.setdefault('organization', {})
            contents['organization']['feeds'] = contents.pop('feeds', [])

    @staticmethod
    def _migrate_schema_100_to_110_organization_repeating_minutes(contents: dict):
        """Migrate 1.0.0 schema to 1.1.0 schema."""
        if (
            contents.get('repeatingMinutes') is not None
            and contents['runtimeLevel'].lower() == 'organization'
        ):
            contents.setdefault('organization', {})
            contents['organization']['repeatingMinutes'] = contents.pop('repeatingMinutes', [])

    @staticmethod
    def _migrate_schema_100_to_110_organization_publish_out_files(contents: dict):
        """Migrate 1.0.0 schema to 1.1.0 schema."""
        if (
            contents.get('publishOutFiles') is not None
            and contents['runtimeLevel'].lower() == 'organization'
        ):
            contents.setdefault('organization', {})
            contents['organization']['publishOutFiles'] = contents.pop('publishOutFiles', [])

    @staticmethod
    def _migrate_schema_100_to_110_input_groups(contents: dict):
        """Migrate 1.0.0 schema to 1.1.0 schema."""
        contents['sections'] = contents.pop('inputGroups', {})
        for section in contents.get('sections'):
            section['sectionName'] = section.pop('group')
            section['params'] = section.pop('inputs')

            # add missing name
            for param in section['params']:
                if param.get('name') is None:
                    param['name'] = param.get('label')

                if 'sequence' in param:
                    del param['sequence']

                if param.get('type') is None:
                    param['type'] = 'String'

    @staticmethod
    def _migrate_schema_100_to_110_notes_per_action(contents: dict):
        """Migrate 1.0.0 schema to 1.1.0 schema."""
        contents['notePerAction'] = contents.pop('notes', {})
        note_per_action = []
        for action, note in contents['notePerAction'].items():
            note_per_action.append({'action': action, 'note': note})
        contents['notePerAction'] = note_per_action

    @staticmethod
    def _migrate_schema_100_to_110_retry(contents: dict):
        """Migrate 1.0.0 schema to 1.1.0 schema."""
        if contents['runtimeLevel'].lower() == 'playbook':
            contents.setdefault('playbook', {})
            if contents.get('playbook', {}).get('retry'):
                contents['playbook']['retry'] = contents.pop('retry', {})

    @staticmethod
    def _migrate_schema_100_to_110_output_groups(contents: dict):
        """Migrate 1.0.0 schema to 1.1.0 schema."""
        outputs = []
        contents['outputData'] = contents.pop('outputGroups', {})
        for display, group in contents.get('outputData').items():
            output_data = {'display': display, 'outputVariables': []}

            # fix schema when output type is assumed
            if isinstance(group, list):
                group = {'String': group}

            for variable_type, variables in group.items():
                for name in variables:
                    disabled = False
                    if name.startswith('~'):
                        name = name.replace('~', '')
                        disabled = True

                    output_data['outputVariables'].append(
                        {
                            'disabled': disabled,
                            'encrypt': False,
                            'intelTypes': [],
                            'name': name,
                            'note': None,
                            'type': variable_type,
                        }
                    )
            outputs.append(output_data)
        contents['outputData'] = outputs

    @staticmethod
    def _migrate_schema_100_to_110_jira_notes(contents: dict):
        """Migrate 1.0.0 schema to 1.1.0 schema."""
        jira_notes = []
        for k, v in contents.get('jira', {}).items():
            # look for the trailer to find our items
            if k == '_TRAILER_':
                for item in v:
                    jira_notes.append(item)
        contents['internalNotes'] = jira_notes

    @staticmethod
    def _migrate_schema_100_to_110_release_notes(contents: dict):
        """Migrate 1.0.0 schema to 1.1.0 schema."""
        release_notes = []
        # need to see if this exist, older apps it might not
        if contents.get('releaseNotes'):
            for k, v in contents.get('releaseNotes').items():
                release_notes.append({'version': k, 'notes': v})
        contents['releaseNotes'] = release_notes

    @cached_property
    def contents(self) -> dict:
        """Return install.json file contents."""

        def _load_contents() -> dict:
            """Load contents from file."""
            contents = {}
            if self.fqfn.is_file():
                try:
                    with self.fqfn.open() as fh:
                        contents = yaml.load(fh, Loader=Loader)  # nosec
                except (OSError, ValueError):  # pragma: no cover
                    self.log.error(
                        f'feature=app-spec-yml, exception=failed-reading-file, filename={self.fqfn}'
                    )
            else:  # pragma: no cover
                self.log.error(
                    f'feature=app-spec-yml, exception=file-not-found, filename={self.fqfn}'
                )

            return contents

        contents = _load_contents()

        # migrate schema from 1.0.0 to 1.1.0
        if contents.get('schemaVersion', '1.0.0') == '1.0.0':
            self._migrate_schema_100_to_110(contents)
        else:
            # reformat file
            self.rewrite_contents(contents)

        # migrate schema
        return _load_contents()

    @staticmethod
    def dict_to_yaml(data: dict) -> str:
        """Convert dict to yaml."""
        return yaml.dump(
            data,
            Dumper=Dumper,
            default_flow_style=False,
            sort_keys=False,
        )

    @property
    def has_spec(self):
        """Return True if App has app_spec.yml file."""
        return self.fqfn.is_file()

    @cached_property
    def model(self) -> 'AppSpecYmlModel':
        """Return the Install JSON model.

        If writing app_spec.yml file after the method then the model will include
        advancedRequest inputs/outputs, etc.
        """
        _contents = self.contents
        # special case for dynamic handling of advancedRequest feature
        if 'advancedRequest' in _contents.get('features', []):
            # look for a Configure section which required for Advanced Request
            if 'Configure' not in [
                section.get('sectionName') for section in _contents.get('sections', [])
            ]:
                raise RuntimeError('The advancedRequest feature requires a Configure section.')

            # Add "Advanced Request" action to Valid Values
            # when "advancedRequest" feature is enabled
            for section in _contents.get('sections', []):
                for param in section.get('params', []):
                    if param.get('name') == 'tc_action' and 'Advanced Request' not in param.get(
                        'validValues', []
                    ):
                        param['validValues'].append('Advanced Request')

                if section.get('sectionName') == 'Configure':
                    section['params'].extend(self._feature_data_advanced_request_inputs)

            # add outputs
            prefix = _contents.get('outputPrefix', '')
            _contents['outputData'].append(self._feature_data_advanced_request_outputs(prefix))

        return AppSpecYmlModel(**self.contents)

    @staticmethod
    def order_data(asy_data: dict) -> dict:
        """Order field with direction given by Business Analysis."""
        asy_data_ordered = {}

        # TODO: [low] rewrite this section to iterate over fields in order

        # "important" fields
        asy_data_ordered['displayName'] = asy_data.get('displayName')
        asy_data_ordered['packageName'] = asy_data.get('packageName')
        asy_data_ordered['appId'] = asy_data.get('appId')
        asy_data_ordered['category'] = asy_data.get('category', '')
        asy_data_ordered['programVersion'] = asy_data.get('programVersion')
        if asy_data.get('displayPath') is not None:
            asy_data_ordered['displayPath'] = asy_data.get('displayPath')
        if asy_data.get('internalNotes') is not None:
            asy_data_ordered['internalNotes'] = asy_data.get('internalNotes', [])
        asy_data_ordered['releaseNotes'] = asy_data.get('releaseNotes', [])
        asy_data_ordered['deprecatesApps'] = asy_data.get('deprecatesApps', [])
        asy_data_ordered['features'] = asy_data.get('features')
        asy_data_ordered['labels'] = asy_data.get('labels', [])
        asy_data_ordered['minServerVersion'] = asy_data.get('minServerVersion')
        asy_data_ordered['note'] = asy_data.get('note') or ''
        if asy_data_ordered.get('serviceDetails'):
            asy_data_ordered['serviceDetails'] = asy_data.get('serviceDetails')
        if asy_data.get('notePerAction') is not None:
            asy_data_ordered['notePerAction'] = asy_data.get('notePerAction')
        asy_data_ordered['runtimeLevel'] = asy_data.get('runtimeLevel')

        # per runtime level (App type)
        if asy_data.get('organization'):
            asy_data_ordered['organization'] = asy_data.get('organization')
        if asy_data.get('playbook'):
            asy_data_ordered['playbook'] = asy_data.get('playbook')
        if asy_data.get('service'):
            asy_data_ordered['service'] = asy_data.get('service')

        # inputs / outputs
        asy_data_ordered['sections'] = asy_data.get('sections')
        if asy_data.get('runtimeLevel').lower() not in (
            'apiservice',
            'feedapiservice',
            'organization',
        ):
            asy_data_ordered['outputData'] = asy_data.get('outputData')
        if asy_data.get('outputPrefix'):
            asy_data_ordered['outputPrefix'] = asy_data.get('outputPrefix')

        # standard fields
        asy_data_ordered['allowOnDemand'] = asy_data.get('allowOnDemand', True)
        if asy_data.get('allowRunAsUser') is not None:
            asy_data_ordered['allowRunAsUser'] = asy_data.get('allowRunAsUser')
        asy_data_ordered['programLanguage'] = asy_data.get('programLanguage')
        asy_data_ordered['languageVersion'] = asy_data.get('languageVersion')
        asy_data_ordered['programMain'] = asy_data.get('programMain')
        asy_data_ordered['listDelimiter'] = asy_data.get('listDelimiter')

        # auto-generated
        asy_data_ordered['schemaVersion'] = str(asy_data.get('schemaVersion', '1.1.0'))

        return asy_data_ordered

    def fix_contents(self, contents: dict) -> dict:
        """Fix missing data"""
        # fix for null appId value
        if 'appId' in contents and contents.get('appId') is None:
            del contents['appId']

        # fix missing packageName
        if contents.get('packageName') is None:
            contents['packageName'] = self.tj.model.package.app_name

        # fix missing outputPrefix
        if contents.get('outputPrefix') is None and 'advancedRequest' in contents.get(
            'features', []
        ):
            contents['outputPrefix'] = self.ij.model.playbook.output_prefix

    def rewrite_contents(self, contents: dict):
        """Rewrite app_spec.yml file."""
        self.fix_contents(contents)

        # exclude_defaults - if False then all unused fields are added in - not good.
        # exclude_none - this should be safe to leave as True.
        # exclude_unset - this should be False to ensure that all fields are included.
        contents = json.loads(
            AppSpecYmlModel(**contents).json(
                by_alias=True,
                exclude_defaults=True,
                exclude_none=True,
                exclude_unset=False,
                sort_keys=True,
            )
        )

        # write the new contents to the file
        self.write(contents)

        return contents

    def write(self, contents: dict):
        """Write yaml to file."""
        with self.fqfn.open(mode='w', encoding='utf-8') as fh:
            fh.write(self.dict_to_yaml(self.order_data(contents)))



"""TcEx Generate Configurations CLI Command"""
# standard library
import os
import shutil
import subprocess  # nosec

# third-party
import black
import isort
import typer
from pydantic import ValidationError

# first-party
from tcex.app_config import AppSpecYml
from tcex.bin.bin_abc import BinABC
from tcex.bin.spec_tool_app_input import SpecToolAppInput
from tcex.bin.spec_tool_app_spec_yml import SpecToolAppSpecYml
from tcex.bin.spec_tool_install_json import SpecToolInstallJson
from tcex.bin.spec_tool_job_json import SpecToolJobJson
from tcex.bin.spec_tool_layout_json import SpecToolLayoutJson
from tcex.bin.spec_tool_readme_md import SpecToolReadmeMd
from tcex.bin.spec_tool_tcex_json import SpecToolTcexJson


class SpecTool(BinABC):
    """Generate App Config Files"""

    def __init__(self, overwrite: bool = False):
        """Initialize class properties."""
        super().__init__()
        self.overwrite = overwrite

        # properties
        self.app_spec_filename = 'app_spec.yml'
        self.asy = AppSpecYml()
        self.report_data = {}
        self.summary_data = []

        # rename app.yaml to app_spec.yml
        self.rename_app_file('app.yaml', 'app_spec.yml')

    def _check_has_spec(self):
        """Validate that the app_spec.yml file exists."""
        if not self.asy.has_spec:
            self.print_failure(
                f'No {self.asy.fqfn.name} file found.\nTry running `tcex spectool '
                f'--app-spec` first to generate the {self.asy.fqfn.name} specification file.'
            )

    @property
    def _git_installed(self) -> bool:
        """Check if git is installed."""
        installed = True
        try:
            #  pylint: disable=consider-using-with
            subprocess.Popen('git', stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)  # nosec
        except OSError:
            installed = False

        self.log.debug(f'action=check-git-installed, results={installed}')
        return installed

    def _git_mv_file(self, src_filename: str, dest_filename: str) -> bool:
        """Check if git is installed."""
        moved = True
        try:
            #  pylint: disable=consider-using-with
            cmd = subprocess.Popen(  # nosec
                ['git', 'mv', src_filename, dest_filename],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            _ = cmd.communicate()[0]
            moved = cmd.returncode == 0
        except OSError:
            moved = False

        self.log.debug(
            f'action=mv-file, src-filename={src_filename}, '
            f'dest-filename={dest_filename}, results={moved}'
        )
        return moved

    def format_code(self, _code):
        """Return formatted code."""

        # run isort on code
        settings_file = None
        if os.path.isfile('setup.cfg'):
            settings_file = 'setup.cfg'
        try:
            isort_config = isort.Config(settings_file=settings_file)
            _code = isort.code(_code, config=isort_config)
        except Exception as ex:
            self.print_failure(f'Formatting of code failed {ex}.')

        # run black formatter on code
        mode = black.FileMode(line_length=100, string_normalization=False)
        try:
            _code = black.format_file_contents(_code, fast=False, mode=mode)
        except black.InvalidInput as ex:
            print(_code)
            self.print_failure(f'Formatting of code failed {ex}.')
        except black.NothingChanged:
            pass

        return _code

    def generate_app_input(self):
        """Generate the app_input.py file."""
        # force migration of app.yaml file
        _ = self.asy.model.app_id

        gen = SpecToolAppInput()
        code = gen.generate()
        self.write_app_file(gen.filename, self.format_code('\n'.join(code)))
        if gen.report_mismatch:
            _report = []
            for r in gen.report_mismatch:
                _report.append(
                    {
                        'label': 'Mismatch',
                        'value': r,
                        'fg_color': typer.colors.YELLOW,
                    }
                )
            self.report_data['Mismatch Report'] = _report

    def generate_app_spec(self, schema: bool):
        """Generate the app_spec.yml file."""
        gen = SpecToolAppSpecYml()
        if schema:
            print(gen.generate_schema())
        else:
            try:
                config = gen.generate()
            except ValidationError as ex:
                self.print_failure(f'Failed Generating app_spec.yml:\n{ex}')

            self.write_app_file(gen.filename, f'{config}\n')

    def generate_install_json(self, schema: bool):
        """Generate the install.json file."""
        gen = SpecToolInstallJson(self.asy)
        if schema:
            print(gen.generate_schema())
        else:
            # check that app_spec.yml exists
            self._check_has_spec()

            try:
                ij = gen.generate()
            except ValidationError as ex:
                self.print_failure(f'Failed Generating install.json:\n{ex}')

            # exclude_defaults - if False then all unused fields are added in - not good.
            # exclude_none - this should be safe to leave as True.
            # exclude_unset - this should be safe to leave as True.
            config = ij.json(
                by_alias=True,
                exclude_defaults=True,
                exclude_none=True,
                exclude_unset=True,
                indent=2,
                sort_keys=True,
            )
            self.write_app_file(gen.filename, f'{config}\n')

    def generate_layout_json(self, schema: bool):
        """Generate the layout.json file."""
        gen = SpecToolLayoutJson(self.asy)
        if schema:
            print(gen.generate_schema())
        elif any([self.ij.model.is_playbook_app, self.ij.model.is_trigger_app]):
            # check that app_spec.yml exists
            self._check_has_spec()

            if self.asy.model.requires_layout:
                try:
                    lj = gen.generate()
                except ValidationError as ex:
                    self.print_failure(f'Failed Generating layout.json:\n{ex}')

                # exclude_defaults - if False then all unused fields are added in - not good.
                # exclude_none - this should be safe to leave as True.
                # exclude_unset - this should be safe to leave as True.
                config = lj.json(
                    by_alias=True,
                    exclude_defaults=True,
                    exclude_none=True,
                    exclude_unset=True,
                    indent=2,
                    sort_keys=True,
                )
                self.write_app_file(gen.filename, f'{config}\n')

    def generate_job_json(self, schema: bool):
        """Generate the job.json file."""
        gen = SpecToolJobJson(self.asy)
        if schema:
            print(gen.generate_schema())
        elif self.asy.model.is_feed_app:
            # check that app_spec.yml exists
            self._check_has_spec()

            try:
                for filename, job in gen.generate():
                    if job is not None:
                        config = job.json(
                            by_alias=True,
                            exclude_defaults=True,
                            exclude_none=True,
                            exclude_unset=True,
                            indent=2,
                            sort_keys=True,
                        )
                        self.write_app_file(filename, f'{config}\n')
            except ValidationError as ex:
                self.print_failure(f'Failed Generating job.json:\n{ex}')

    def generate_readme_md(self):
        """Generate the README.me file."""
        # check that app_spec.yml exists
        self._check_has_spec()

        gen = SpecToolReadmeMd(self.asy)
        readme_md = gen.generate()
        self.write_app_file(gen.filename, '\n'.join(readme_md))

    def generate_tcex_json(self, schema: bool):
        """Generate the tcex.json file."""
        gen = SpecToolTcexJson(self.asy)
        if schema:
            print(gen.generate_schema())
        else:
            # check that app_spec.yml exists
            self._check_has_spec()

            try:
                tj = gen.generate()
            except ValidationError as ex:
                self.print_failure(f'Failed Generating tcex.json:\n{ex}')

            # exclude_defaults - if False then all unused fields are added in - not good.
            # exclude_none - this should be safe to leave as True.
            # exclude_unset - this should be safe to leave as True.
            config = tj.json(
                by_alias=True,
                exclude_defaults=True,
                exclude_none=True,
                exclude_unset=True,
                indent=2,
                sort_keys=True,
            )
            self.write_app_file(gen.filename, f'{config}\n')

    def rename_app_file(self, src_filename: str, dest_filename: str):
        """Rename the app.yaml file to app_spec.yml."""
        if os.path.isfile(src_filename):
            self.log.debug(
                f'action=rename-file, src-filename={src_filename}, dest-filename={dest_filename}'
            )

            moved = False
            if self._git_installed is True:
                # to not loose git history of the file
                moved = self._git_mv_file(src_filename, dest_filename)

            if moved is False:
                shutil.move('app.yaml', 'app_spec.yml')

    def print_report(self):
        """Print report data."""
        for report_name, report_data in self.report_data.items():
            self.print_title(report_name)
            for report in report_data:
                self.print_setting(
                    report.get('label'), report.get('value'), fg_color=report.get('fg_color')
                )

    def print_summary(self):
        """Print action summary."""
        if self.summary_data:
            self.print_title('Summary Data')
            for summary in self.summary_data:
                self.print_setting(
                    summary.get('label'), summary.get('value'), fg_color=summary.get('fg_color')
                )

    def write_app_file(self, file_name: str, contents: str):
        """Write contents to file."""
        action = 'Created'
        write_file = True
        if os.path.isfile(file_name):
            action = 'Updated'
            if self.overwrite is False:
                prompt = self.prompt_choice(
                    f'The {file_name} file already exists: Overwrite?',
                    choices=['yes', 'no'],
                    default='no',
                    fg_color=typer.colors.YELLOW,
                )
                if prompt == 'no':
                    action = 'Skipped'
                    write_file = False

        if write_file is True:
            with open(file_name, 'w') as f:
                f.write(contents)

        self.summary_data.append(
            {
                'fg_color': typer.colors.GREEN,
                'label': f'{action} File',
                'value': file_name,
            }
        )



"""Service module for TcEx Framework"""
# flake8: noqa
# first-party
from tcex.services.api_service import ApiService
from tcex.services.common_service_trigger import CommonServiceTrigger
from tcex.services.mqtt_message_broker import MqttMessageBroker
from tcex.services.webhook_trigger_service import WebhookTriggerService



"""Indicator Entity Field (Model) Type"""
# standard library
from typing import TYPE_CHECKING, List

# third-party
from pydantic import validator

# first-party
from tcex.input.field_types.exception import InvalidEmptyValue, InvalidEntityType
from tcex.input.field_types.tc_entity import TCEntity

if TYPE_CHECKING:  # pragma: no cover
    # third-party
    from pydantic.fields import ModelField

CASE_MANAGEMENT_TYPES = [
    'artifact',
    'case',
    'task',
    'artifact type',
    'notes',
    'workflow event',
    'workflow template',
]


# pylint: disable=no-self-argument, no-self-use
class CaseManagementEntity(TCEntity):
    """Case Management Entity Field (Model) Type"""

    @validator('type')
    def is_empty(cls, value: str, field: 'ModelField') -> str:
        """Validate that the value is a non-empty string."""
        if isinstance(value, str) and value.replace(' ', '') == '':
            raise InvalidEmptyValue(field_name=field.name)
        return value

    @validator('type')
    def is_type(cls, value: str, field: 'ModelField') -> str:
        """Validate that the entity is of Indicator type."""
        if value.lower() not in CASE_MANAGEMENT_TYPES:
            raise InvalidEntityType(
                field_name=field.name, entity_type='CaseManagement', value=value
            )
        return value


def case_management_entity(case_management_types: List[str] = None) -> type:
    """Return custom model for Case Management Entity."""

    class CustomCaseManagementEntity(CaseManagementEntity):
        """Indicator Entity Field (Model) Type"""

        @validator('type', allow_reuse=True)
        def is_empty(cls, value: str, field: 'ModelField') -> str:
            """Validate that the value is a non-empty string."""
            if isinstance(value, str) and value.replace(' ', '') == '':
                raise InvalidEmptyValue(field_name=field.name)
            return value

        @validator('type', allow_reuse=True)
        def is_type(cls, value: str, field: 'ModelField') -> str:
            """Validate that the entity is of a specific Indicator type."""
            if value.lower() not in [i.lower() for i in case_management_types]:
                raise InvalidEntityType(
                    field_name=field.name, entity_type=str(case_management_types), value=value
                )
            return value

    return CustomCaseManagementEntity


ArtifactEntity: CaseManagementEntity = case_management_entity(case_management_types=['Artifact'])
CaseEntity: CaseManagementEntity = case_management_entity(case_management_types=['Case'])
TaskEntity: CaseManagementEntity = case_management_entity(case_management_types=['Task'])
ArtifactTypeEntity: CaseManagementEntity = case_management_entity(
    case_management_types=['Artifact Type']
)
NoteEntity: CaseManagementEntity = case_management_entity(case_management_types=['Note'])
WorkflowEventEntity: CaseManagementEntity = case_management_entity(
    case_management_types=['Workflow Event']
)
WorkflowTemplateEntity: CaseManagementEntity = case_management_entity(
    case_management_types=['Workflow Template']
)



"""String Field Type"""
# standard library
import re
from typing import TYPE_CHECKING, Callable, Optional, Union

# first-party
from tcex.input.field_types.exception import (
    InvalidEmptyValue,
    InvalidLengthValue,
    InvalidPatternValue,
    InvalidType,
    InvalidVariableType,
)

if TYPE_CHECKING:  # pragma: no cover
    # third-party
    from pydantic.fields import ModelField

    # first-party
    from tcex.utils.variables import StringVariable


class String(str):
    """String Field Type"""

    allow_empty: bool = True
    max_length: Optional[int] = None
    min_length: Optional[int] = None
    regex: Optional[str] = None
    strip: bool = False

    @classmethod
    def __get_validators__(cls) -> Callable:
        """Run validators / modifiers on input."""
        yield cls.validate_variable_type
        yield cls.validate_type
        yield cls.validate_strip
        yield cls.validate_allow_empty
        yield cls.validate_max_length
        yield cls.validate_min_length
        yield cls.validate_regex

    @classmethod
    def validate_allow_empty(cls, value: Union[str, 'StringVariable'], field: 'ModelField') -> str:
        """Raise exception if value is empty and allow_empty is False."""
        if cls.allow_empty is False:
            if isinstance(value, str) and value == '':
                raise InvalidEmptyValue(field_name=field.name)

        return value

    @classmethod
    def validate_max_length(cls, value: Union[str, 'StringVariable'], field: 'ModelField') -> str:
        """Raise exception if value does not match pattern."""
        if cls.max_length is not None and len(value) > cls.max_length:
            raise InvalidLengthValue(
                field_name=field.name, constraint=cls.max_length, operation='max'
            )
        return value

    @classmethod
    def validate_min_length(cls, value: Union[str, 'StringVariable'], field: 'ModelField') -> str:
        """Raise exception if value does not match pattern."""
        if cls.min_length is not None and len(value) < cls.min_length:
            raise InvalidLengthValue(
                field_name=field.name, constraint=cls.min_length, operation='min'
            )
        return value

    @classmethod
    def validate_regex(cls, value: Union[str, 'StringVariable'], field: 'ModelField') -> str:
        """Raise exception if value does not match pattern."""
        if isinstance(cls.regex, str):
            if not re.compile(cls.regex).match(value):
                raise InvalidPatternValue(field_name=field.name, pattern=cls.regex)
        return value

    @classmethod
    def validate_strip(cls, value: Union[bytes, 'StringVariable']) -> bytes:
        """Raise exception if value is not a Binary type."""
        if cls.strip is True:
            value = value.strip()
        return value

    @classmethod
    def validate_type(cls, value: Union[str, 'StringVariable'], field: 'ModelField') -> str:
        """Raise exception if value is not a String type."""
        if not isinstance(value, str):
            raise InvalidType(
                field_name=field.name, expected_types='(str)', provided_type=type(value)
            )
        return value

    @classmethod
    def validate_variable_type(
        cls, value: Union[str, 'StringVariable'], field: 'ModelField'
    ) -> str:
        """Raise exception if value is not a String type."""
        if hasattr(value, '_variable_type') and value._variable_type != 'String':
            raise InvalidVariableType(
                field_name=field.name, expected_type='String', provided_type=value._variable_type
            )
        return value


def string(
    allow_empty: bool = True,
    max_length: Optional[int] = None,
    min_length: Optional[int] = None,
    regex: Optional[str] = None,
    strip: bool = False,
) -> type:
    """Return configured instance of String."""
    namespace = dict(
        allow_empty=allow_empty,
        max_length=max_length,
        min_length=min_length,
        regex=regex,
        strip=strip,
    )
    return type('ConstrainedString', (String,), namespace)



"""Create Config Model"""
# pylint: disable=no-self-argument,no-self-use
# standard library
from typing import Any, Dict, List

# third-party
from pydantic import BaseModel, root_validator, validator

# first-party
from tcex.app_config import InstallJson

# get instance of InstallJson
ij = InstallJson()


class CreateConfigModel(BaseModel):
    """Create Config Model"""

    tc_playbook_out_variables: List[str]
    trigger_id: int

    @validator('tc_playbook_out_variables', pre=True)
    def _tc_playbook_out_variables(cls, v):
        """Convert tc_playbook_out_variables into a list of strings.

        This value comes-in as a comma-separated list.
        """
        return v.split(',') if v else []

    # TODO: [low] workaround for PLAT-4393
    @root_validator(pre=True)
    def _update_inputs(cls, values: Dict[str, Any]):
        """Convert empty strings to None.

        Workarounds for core issues:
            - Core sends '' for field that are not populated instead of sending a null value.
            - Core sends a string for multi-value inputs instead of an array.
        """
        for field, value in values.items():
            param = ij.model.get_param(field)
            if param.type is not None and (
                param.type.lower() == 'multichoice' or param.allow_multiple
            ):
                if value is not None and not isinstance(value, list):
                    values[field] = value.split(ij.model.list_delimiter or '|')

            if value == '':
                values[field] = None

        return values



"""ThreatConnect TI Email"""
# standard library
from typing import TYPE_CHECKING

# first-party
from tcex.api.tc.v2.threat_intelligence.mappings.group.group import Group

if TYPE_CHECKING:
    # first-party
    from tcex.api.tc.v2.threat_intelligence.threat_intelligence import ThreatIntelligence


class Vulnerability(Group):
    """Unique API calls for Vulnerability API Endpoints

    Args:
        ti (ThreatIntelligence): An instance of the ThreatIntelligence Class.
        name (str, kwargs): [Required for Create] The name for this Group.
        owner (str, kwargs): The name for this Group. Default to default Org when not provided
    """

    def __init__(self, ti: 'ThreatIntelligence', **kwargs):
        """Initialize Class properties."""
        super().__init__(
            ti,
            sub_type='Vulnerability',
            api_entity='vulnerability',
            api_branch='vulnerabilities',
            **kwargs
        )



"""WorkflowTemplate / WorkflowTemplates Object"""
# standard library
from typing import Union

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.object_abc import ObjectABC
from tcex.api.tc.v3.object_collection_abc import ObjectCollectionABC
from tcex.api.tc.v3.workflow_templates.workflow_template_filter import WorkflowTemplateFilter
from tcex.api.tc.v3.workflow_templates.workflow_template_model import (
    WorkflowTemplateModel,
    WorkflowTemplatesModel,
)


class WorkflowTemplates(ObjectCollectionABC):
    """WorkflowTemplates Collection.

    # Example of params input
    {
        'result_limit': 100,  # Limit the retrieved results.
        'result_start': 10,  # Starting count used for pagination.
        'fields': ['caseId', 'summary']  # Select additional return fields.
    }

    Args:
        session (Session): Session object configured with TC API Auth.
        tql_filters (list): List of TQL filters.
        params (dict): Additional query params (see example above).
    """

    def __init__(self, **kwargs):
        """Initialize class properties."""
        super().__init__(
            kwargs.pop('session', None), kwargs.pop('tql_filter', None), kwargs.pop('params', None)
        )
        self._model = WorkflowTemplatesModel(**kwargs)
        self.type_ = 'workflow_templates'

    def __iter__(self) -> 'WorkflowTemplate':
        """Iterate over CM objects."""
        return self.iterate(base_class=WorkflowTemplate)

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.WORKFLOW_TEMPLATES.value

    @property
    def filter(self) -> 'WorkflowTemplateFilter':
        """Return the type specific filter object."""
        return WorkflowTemplateFilter(self.tql)


class WorkflowTemplate(ObjectABC):
    """WorkflowTemplates Object.

    Args:
        config_attribute (None, kwargs): The **config attribute** for the Workflow_Template.
        description (str, kwargs): The **description** for the Workflow_Template.
        name (str, kwargs): The **name** for the Workflow_Template.
        version (int, kwargs): The **version** for the Workflow_Template.
    """

    def __init__(self, **kwargs):
        """Initialize class properties."""
        super().__init__(kwargs.pop('session', None))

        # properties
        self._model = WorkflowTemplateModel(**kwargs)
        self._nested_field_name = 'workflowTemplates'
        self._nested_filter = 'has_workflow_template'
        self.type_ = 'Workflow Template'

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.WORKFLOW_TEMPLATES.value

    @property
    def model(self) -> 'WorkflowTemplateModel':
        """Return the model data."""
        return self._model

    @model.setter
    def model(self, data: Union['WorkflowTemplateModel', dict]):
        """Create model using the provided data."""
        if isinstance(data, type(self.model)):
            # provided data is already a model, nothing required to change
            self._model = data
        elif isinstance(data, dict):
            # provided data is raw response, load the model
            self._model = type(self.model)(**data)
        else:
            raise RuntimeError(f'Invalid data type: {type(data)} provided.')

    @property
    def as_entity(self) -> dict:
        """Return the entity representation of the object."""
        type_ = self.type_

        return {'type': type_, 'id': self.model.id, 'value': self.model.name}



"""Case Management Filter Abstract Base Class"""
# standard library
from abc import ABC
from typing import TYPE_CHECKING, List

# first-party
from tcex.utils.utils import Utils

if TYPE_CHECKING:
    # first-party
    from tcex.api.tc.v3.tql.tql import Tql


class FilterABC(ABC):
    """Case Management Filter Abstract Base Class"""

    def __init__(self, tql: 'Tql'):
        """Initialize Class properties"""
        self._tql = tql

        # properties
        self.utils = Utils()

    @property
    def _api_endpoint(self):
        raise NotImplementedError('Child class must implement this method.')

    @property
    def implemented_keywords(self) -> List[str]:
        """Return implemented TQL keywords."""
        keywords = []
        for prop in dir(self):
            if prop.startswith('_') or prop in ['tql']:
                continue
            keywords.append(prop)

        return keywords

    @property
    def tql(self) -> 'Tql':
        """Return the current TQL instance."""
        return self._tql

    @tql.setter
    def tql(self, tql: str):
        """Filter objects based on TQL expression.

        Args:
            tql: The raw TQL string for the filter.
        """
        self.tql.set_raw_tql(tql)

    def __str__(self) -> str:
        """Return string representation of TQL."""
        return self.tql.raw_tql or self.tql.as_str



"""Group Entity Field (Model) Type"""
# standard library
from typing import TYPE_CHECKING, Dict, List

# third-party
from pydantic import validator

# first-party
from tcex.api.tc.utils.threat_intel_utils import ThreatIntelUtils
from tcex.input.field_types.exception import InvalidEmptyValue, InvalidEntityType
from tcex.input.field_types.tc_entity import TCEntity
from tcex.pleb.registry import registry

if TYPE_CHECKING:  # pragma: no cover
    # third-party
    from pydantic.fields import ModelField


# pylint: disable=no-self-argument, no-self-use
class GroupEntity(TCEntity):
    """Group Entity Field (Model) Type"""

    @validator('type')
    def is_type(cls, value: str, field: 'ModelField') -> str:
        """Validate that the value is a non-empty string.

        Without the always and pre args, None values will validated before this validator is called.
        """
        if isinstance(value, str) and value.replace(' ', '') == '':
            raise InvalidEmptyValue(field_name=field.name)

        ti_utils = ThreatIntelUtils(session_tc=registry.session_tc)
        if value not in ti_utils.group_types:
            raise InvalidEntityType(field_name=field.name, entity_type='Group', value=value)
        return value


def group_entity(group_types: List[str] = None) -> type:
    """Return custom model for Group Entity."""

    class CustomGroupEntity(GroupEntity):
        """Group Entity Field (Model) Type"""

        @validator('type', allow_reuse=True)
        def is_empty(cls, value: str, field: 'ModelField') -> Dict[str, str]:
            """Validate that the value is a non-empty string."""
            if isinstance(value, str) and value.replace(' ', '') == '':
                raise InvalidEmptyValue(field_name=field.name)
            return value

        @validator('type', allow_reuse=True)
        def is_type(cls, value: str, field: 'ModelField') -> Dict[str, str]:
            """Validate that the entity is of a specific Group type."""
            if value.lower() not in [i.lower() for i in group_types]:
                raise InvalidEntityType(
                    field_name=field.name, entity_type=str(group_types), value=value
                )
            return value

    return CustomGroupEntity


AdversaryEntity: GroupEntity = group_entity(group_types=['Adversary'])
CampaignEntity: GroupEntity = group_entity(group_types=['Campaign'])
DocumentEntity: GroupEntity = group_entity(group_types=['Document'])
EmailEntity: GroupEntity = group_entity(group_types=['Email'])
EventEntity: GroupEntity = group_entity(group_types=['Event'])
IncidentEntity: GroupEntity = group_entity(group_types=['Incident'])
IntrusionSetEntity: GroupEntity = group_entity(group_types=['Intrusion Set'])
SignatureEntity: GroupEntity = group_entity(group_types=['Signature'])
ReportEntity: GroupEntity = group_entity(group_types=['Report'])
ThreatEntity: GroupEntity = group_entity(group_types=['Threat'])
TaskEntity: GroupEntity = group_entity(group_types=['Task'])



"""TcEx Generate Configurations CLI Command"""
# standard library
from typing import TYPE_CHECKING

# first-party
from tcex.app_config.models import LayoutJsonModel
from tcex.bin.bin_abc import BinABC

if TYPE_CHECKING:
    # first-party
    from tcex.app_config import AppSpecYml


class SpecToolLayoutJson(BinABC):
    """Generate App Config File"""

    def __init__(self, asy: 'AppSpecYml'):
        """Initialize class properties."""
        super().__init__()
        self.asy = asy

        # properties
        self.filename = 'layout.json'

    def generate(self):
        """Generate the layout.json file data."""

        layout_json_data = {
            'inputs': self.asy.model.inputs,
            'outputs': self.asy.model.outputs,
        }
        return LayoutJsonModel(**layout_json_data)

    @staticmethod
    def generate_schema():
        """Return the schema for the install.json file."""
        return LayoutJsonModel.schema_json(indent=2, sort_keys=True)
